/*
 * Copyright (C) 2005-2006 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file audiod.c the paraslash's audio daemon */

#include <sys/time.h> /* gettimeofday */
#include "para.h"

#include "audiod.cmdline.h"
#include "list.h"
#include "close_on_fork.h"
#include "recv.h"
#include "filter.h"
#include "grab_client.cmdline.h"
#include "grab_client.h"
#include "ringbuffer.h"

#include "error.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"

/** define the array of error lists needed by para_audiod */
INIT_AUDIOD_ERRLISTS;
/** define the array containing all supported audio formats */
DEFINE_AUDIO_FORMAT_ARRAY;

/**
 * the possible modes of operation
 *
 * - off: disconnect from para_server
 * - on: receive status information from para_server and play the audio stream
 * - sb: only receive status information but not the audio stream
*/
enum {AUDIOD_OFF, AUDIOD_ON, AUDIOD_STANDBY};

/** defines how to handle one supported audio format */
struct audio_format_info {
/** pointer to the receiver for this audio format */
	struct receiver *receiver;
/** the receiver configuration */
	void *receiver_conf;
/** the number of filters that should be activated for this audio format */
	unsigned int num_filters;
/** pointer to the array of filters to be activated */
	struct filter **filters;
/** pointer to the array of filter configurations */
	void **filter_conf;
/** output of the last filter is written to stdin of this command */
	char *write_cmd;
/** do not start receiver/filters/writer before this time */
	struct timeval restart_barrier;
};

/**
 * describes one instance of a receiver-filter-writer chain
 *
 * \sa receier_node, receiver, filter, filter_node, filter_chain_info
  */
struct slot_info {
/** number of the audio format in this slot */
	int format;
/** the file descriptor of the writer */
	int write_fd;
/** the process id of the writer */
	pid_t wpid;
/** time of the last successful read from the receiver */
	struct timeval rtime;
/** time the last write to the write fd happend */
	struct timeval wtime;
/** writer start time */
	struct timeval wstime;
/** did we include \a write_fd in the fdset */
	int  wcheck;
/** set to one if we have sent the TERM signal to \a wpid */
	int wkilled;
/** the receiver info associated with this slot */
	struct receiver_node *receiver_node;
/** the active filter chain */
	struct filter_chain_info *fci;
};

static struct slot_info slot[MAX_STREAM_SLOTS];

/** defines one command of para_audiod */
struct audiod_command {
/** the name of the command */
const char *name;
/** pointer to the function that handles the command */
int (*handler)(int, int, char**);
/** one-line description of the command */
const char *description;
/** summary of the command line options */
const char *synopsis;
/** the long help text */
const char *help;
};

extern const char *status_item_list[NUM_STAT_ITEMS];

static int com_grab(int, int, char **);
static int com_cycle(int, int, char **);
static int com_help(int, int, char **);
static int com_off(int, int, char **);
static int com_on(int, int, char **);
static int com_sb(int, int, char **);
static int com_stat(int, int, char **);
static int com_term(int, int, char **);
static int stat_pipe = -1, signal_pipe;

static struct gengetopt_args_info conf;
static struct timeval server_stream_start, sa_time_diff;
static int playing, current_decoder = -1,
	audiod_status = AUDIOD_ON, offset_seconds, length_seconds,
	sa_time_diff_sign = 1, audiod_socket = -1;
static char *af_status, /* the audio format announced in server status */
	*socket_name, *hostname;
/** how many status items to remember */
#define RINGBUFFER_SIZE 32
static void *stat_item_ringbuf;
static FILE *logfile;
static const struct timeval restart_delay = {0, 300 * 1000};

static struct audio_format_info afi[] = {

[AUDIO_FORMAT_MP3] =
	{
		.write_cmd = "para_play",
	},
[AUDIO_FORMAT_OGG] =
	{
		.write_cmd = "para_play",
	},
};

static struct audiod_command cmds[] = {
{
.name = "cycle",
.handler = com_cycle,
.description = "switch to next mode",
.synopsis = "cycle",
.help =

"on -> standby -> off -> on\n"

},
{
.name = "grab",
.handler = com_grab,
.description = "grab the audio stream",
.synopsis = "-- grab [grab_options]",
.help =
"grab ('splice') the audio stream at any position in the filter      \n"
"chain and send that data back to the client. \n"
"Available options:\n\n"
GRAB_HELP_TXT
},
{
.name = "help",
.handler = com_help,
.description = "display command list or help for given command",
.synopsis = "help [command]",
.help =

"When I was younger, so much younger than today, I never needed\n"
"anybody's help in any way. But now these days are gone, I'm not so\n"
"self assured. Now I find I've changed my mind and opened up the doors.\n"
"\n"
"				-- Beatles: Help\n"

},
{
.name = "off",
.handler = com_off,
.description = "deactivate para_audiod",
.synopsis = "off",
.help =

"Close connection to para_server and stop all decoders.\n"

},
{
.name = "on",
.handler = com_on,
.description = "activate para_audiod",
.synopsis = "on",
.help =

"Establish connection to para_server, retrieve para_server's current\n"
"status. If playing, start corresponding decoder. Otherwise stop\n"
"all decoders.\n"

},
{
.name = "sb",
.handler = com_sb,
.description = "enter standby mode",
.synopsis = "sb",
.help =

"Stop all decoders but leave connection to para_server open.\n"

},
{
.name = "stat",
.handler = com_stat,
.description = "print status information",
.synopsis = "stat",
.help =

"Add para_audiod status information to para_server's status information\n"
"and dump everything to stdout.\n"

},
{
.name = "term",
.handler = com_term,
.description = "terminate audiod",
.synopsis = "term",
.help =

"Stop all decoders, shut down connection to para_server and exit.\n"

},
{
.name = NULL,
}
};

/** iterate over all slots */
#define FOR_EACH_SLOT(slot) for (slot = 0; slot < MAX_STREAM_SLOTS; slot++)
/** iterate over all supported audio formats */
#define FOR_EACH_AUDIO_FORMAT(af) for (af = 0; af < NUM_AUDIO_FORMATS; af++)
/** iterate over the array of all audiod commands */
#define FOR_EACH_COMMAND(c) for (c = 0; cmds[c].name; c++)

/**
 * get the audio format number
 * \param name the name of the audio format
 *
 * \return The audio format number on success, -E_UNSUPPORTED_AUDIO_FORMAT if
 * \a name is not a supported audio format.
 */
int get_audio_format_num(char *name)
{
	int i;
	FOR_EACH_AUDIO_FORMAT(i)
		if (!strcmp(name, audio_formats[i]))
			return i;
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

/*
 * log function. first argument is loglevel.
 */
void para_log(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";

	if (ll < conf.loglevel_arg)
		return;
	if (!logfile && conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	if (!logfile && conf.daemon_given)
		return;
	if (!logfile) {
		if (ll < WARNING)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s %s ", str, hostname);
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "%i ", ll);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static int client_write(int fd, const char *buf)
{
	size_t len = strlen(buf);
	return write(fd, buf, len) != len? -E_CLIENT_WRITE: 1;
}

static char *get_time_string(struct timeval *newest_stime)
{
	struct timeval now, diff, adj_stream_start, tmp;
	int total = 0, use_server_time = 1;

	if (!playing)
		return make_message("%s:", length_seconds?
			"" : status_item_list[SI_PLAY_TIME]);
	if (audiod_status == AUDIOD_OFF)
		goto out;
	if (sa_time_diff_sign > 0)
		tv_diff(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	else
		tv_add(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	tmp = adj_stream_start;
	if (newest_stime && audiod_status == AUDIOD_ON) {
		tv_diff(newest_stime, &adj_stream_start, &diff);
		if (tv2ms(&diff) < 5000) {
			tmp = *newest_stime;
			use_server_time = 0;
		}
	}
	gettimeofday(&now, NULL);
	tv_diff(&now, &tmp, &diff);
	total = diff.tv_sec + offset_seconds;
	if (total > length_seconds)
		total = length_seconds;
	if (total < 0)
		total = 0;
out:
	return make_message(
		"%s:%s%d:%02d [%d:%02d] (%d%%/%d:%02d)",
		status_item_list[SI_PLAY_TIME],
		use_server_time? "~" : "",
		total / 60,
		total % 60,
		(length_seconds - total) / 60,
		(length_seconds - total) % 60,
		length_seconds? (total * 100 + length_seconds / 2) /
			length_seconds : 0,
		length_seconds / 60,
		length_seconds % 60
	);
}

static char *audiod_status_string(void)
{
	int i;
	struct timeval *newest_stime = NULL;
	char *ret, *time_string, *uptime_string, *decoder_flags =
		para_malloc((MAX_STREAM_SLOTS + 1) * sizeof(char));

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		char flag = '0';
		if (s->receiver_node)
			flag += 1;
		if (s->wpid > 0)
			flag += 2;
		if (flag != '0')
			flag += s->format * 4;
		decoder_flags[i] = flag;
		if (s->wpid <= 0)
			continue;
		if (newest_stime && tv_diff(&s->wstime, newest_stime, NULL) <= 0)
			continue;
		newest_stime = &s->wstime;
	}
	decoder_flags[MAX_STREAM_SLOTS] = '\0';
	time_string = get_time_string(newest_stime);
	uptime_string = uptime_str();
	ret = make_message("%s:%s\n%s:%s\n%s:%s\n%s",
		status_item_list[SI_AUDIOD_UPTIME], uptime_string,
		status_item_list[SI_DECODER_FLAGS], decoder_flags,
		status_item_list[SI_AUDIOD_STATUS], audiod_status == AUDIOD_ON?
			"on" : (audiod_status == AUDIOD_OFF? "off": "sb"),
		time_string);
	free(uptime_string);
	free(decoder_flags);
	free(time_string);
	return ret;
}

static char *configfile_exists(void)
{
	static char *config_file;

	if (!config_file) {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	return file_exists(config_file)? config_file : NULL;
}

static void setup_signal_handling(void)
{
	signal_pipe = para_signal_init();
	PARA_INFO_LOG("signal pipe: fd %d\n", signal_pipe);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGCHLD);
	para_install_sighandler(SIGHUP);
	signal(SIGPIPE, SIG_IGN);
}

static void audiod_status_dump(void)
{
	static char *prev_status;
	char *tmp = audiod_status_string();

	if (!prev_status || strcmp(tmp, prev_status))
		stat_client_write(tmp);
	free(prev_status);
	prev_status = tmp;
}

static void clear_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	PARA_INFO_LOG("clearing slot %d\n", slot_num);
	memset(s, 0, sizeof(struct slot_info));
	s->format = -1;
}

static void kill_stream_writer(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	if (s->format < 0 || s->wkilled || s->wpid <= 0)
		return;
	PARA_DEBUG_LOG("kill -TERM %d (%s stream writer in slot %d)\n",
		s->wpid, audio_formats[s->format], slot_num);
	kill(s->wpid, SIGTERM);
	s->wkilled = 1;
	s->fci->error = 1;
}

static void close_receiver(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a;
	struct timeval now;

	if (s->format < 0 || !s->receiver_node)
		return;
	a = &afi[s->format];
	PARA_NOTICE_LOG("closing %s recevier in slot %d\n",
		audio_formats[s->format] , slot_num);
	a->receiver->close(s->receiver_node);
	free(s->receiver_node);
	s->receiver_node = NULL;
	gettimeofday(&now, NULL);
	tv_add(&now, &restart_delay, &a->restart_barrier); /* FIXME: Use set_restart_barrier() */
}

static void kill_all_decoders(void)
{
	int i;

	FOR_EACH_SLOT(i)
		if (slot[i].format >= 0) {
			PARA_INFO_LOG("stopping decoder in slot %d\n", i);
			kill_stream_writer(i);
		}
}

static void set_restart_barrier(int format, struct timeval *now)
{
	struct timeval tmp;

	if (now)
		tmp = *now;
	else
		gettimeofday(&tmp, NULL);
	tv_add(&tmp, &restart_delay, &afi[format].restart_barrier);
}

static void check_sigchld(void)
{
	pid_t pid;
	int i;
	struct timeval now;
	gettimeofday(&now, NULL);

reap_next_child:
	pid = para_reap_child();
	if (pid <= 0)
		return;
	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		long lifetime;
		if (s->format < 0)
			continue;
		if (pid == s->wpid) {
			s->wpid = -1;
			lifetime = now.tv_sec - s->wstime.tv_sec;
			PARA_INFO_LOG("%s stream writer in slot %d died "
				"after %li secs\n",
				audio_formats[s->format], i, lifetime);
			set_restart_barrier(s->format, &now);
			goto reap_next_child;
		}
	}
	PARA_CRIT_LOG("para_client died (pid %d)\n", pid);
	goto reap_next_child;
}

static int get_empty_slot(void)
{
	int i;
	struct slot_info *s;

	FOR_EACH_SLOT(i) {
		s = &slot[i];
		if (s->format < 0) {
			clear_slot(i);
			return i;
		}
		if (s->write_fd > 0 || s->wpid > 0)
			continue;
		if (s->receiver_node)
			continue;
		if (s->fci)
			continue;
		clear_slot(i);
		return i;
	}
	return -E_NO_MORE_SLOTS;
}

static int decoder_running(int format)
{
	int i, ret = 0;
	struct slot_info *s;

	FOR_EACH_SLOT(i) {
		s = &slot[i];
		if (s->format == format && s->receiver_node)
			ret |= 1;
		if (s->format == format && s->wpid > 0)
			ret |= 2;
	}
	return ret;
}

static void close_stat_pipe(void)
{
	char *msg;
	int i;

	if (stat_pipe < 0)
		return;
	PARA_NOTICE_LOG("%s", "closing status pipe\n");
	close(stat_pipe);
	del_close_on_fork_list(stat_pipe);
	stat_pipe = -1;
	kill_all_decoders();
	for (i = 0; i < RINGBUFFER_SIZE; i++)
		ringbuffer_add(stat_item_ringbuf, para_strdup(NULL));
	dump_empty_status();
	length_seconds = 0;
	offset_seconds = 0;
	audiod_status_dump();
	playing = 0;
	msg = make_message("%s:no connection to para_server\n",
		status_item_list[SI_STATUS_BAR]);
	ringbuffer_add(stat_item_ringbuf, msg);
	stat_client_write(msg);
	free(msg);
}

static void __noreturn clean_exit(int status, const char *msg)
{
	PARA_EMERG_LOG("%s\n", msg);
	kill_all_decoders();
	if (socket_name)
		unlink(socket_name);
	if (stat_pipe >= 0)
		close_stat_pipe();
	exit(status);
}

static char *glob_cmd(char *cmd)
{
	char *ret, *replacement;
	struct timeval tmp, delay, rss; /* real stream start */

	delay.tv_sec = conf.stream_delay_arg / 1000;
	delay.tv_usec = (conf.stream_delay_arg % 1000) * 1000;
//	PARA_INFO_LOG("delay: %lu:%lu\n", delay.tv_sec, delay.tv_usec);
	if (sa_time_diff_sign < 0)
		tv_add(&server_stream_start, &sa_time_diff, &rss);
	else
		tv_diff(&server_stream_start, &sa_time_diff, &rss);
	tv_add(&rss, &delay, &tmp);
	replacement = make_message("%lu:%lu", tmp.tv_sec, tmp.tv_usec);
	ret = s_a_r(cmd, "STREAM_START", replacement);
	free(replacement);
	if (!ret)
		goto out;
	PARA_INFO_LOG("cmd: %s, repl: %s\n", cmd, ret);
	{
	struct timeval now;
	gettimeofday(&now, NULL);
	PARA_INFO_LOG("now: %lu:%lu\n", now.tv_sec, now.tv_usec);
	}
out:
	return ret;
}

/** get the number of filters for the given audio format */
int num_filters(int audio_format_num)
{
	return afi[audio_format_num].num_filters;
}

static void open_filters(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];
	int nf = a->num_filters;
	int i;

	s->fci = para_calloc(sizeof(struct filter_chain_info));
	INIT_LIST_HEAD(&s->fci->filters);
	if (!nf)
		return;
	s->fci->inbuf = s->receiver_node->buf;
	s->fci->in_loaded = &s->receiver_node->loaded;
	s->fci->outbuf = s->receiver_node->buf;
	s->fci->out_loaded = &s->receiver_node->loaded;
	s->fci->eof = &s->receiver_node->eof;
	for (i = 0; i < nf; i++) {
		struct filter_node *fn = para_calloc(sizeof(struct filter_node));
		fn->conf = a->filter_conf[i];
		fn->fci = s->fci;
		fn->filter = a->filters[i];
		INIT_LIST_HEAD(&fn->callbacks);
		list_add_tail(&fn->node, &s->fci->filters);
		fn->filter->open(fn);
		PARA_NOTICE_LOG("%s filter %d/%d (%s) started in slot %d\n",
			audio_formats[s->format], i + 1,  nf,
			fn->filter->name, slot_num);
		s->fci->outbuf = fn->buf;
		s->fci->out_loaded = &fn->loaded;
	}
	PARA_DEBUG_LOG("output buffer for filter chain %p: %p\n", s->fci,
		s->fci->outbuf);
}

static struct filter_node *find_filter_node(int slot_num, int format, int filternum)
{
	struct filter_node *fn;
	int i, j;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		if (s->format < 0 || !s->fci)
			continue;
		if (slot_num >= 0 && slot_num != i)
			continue;
		if (format >= 0 && s->format != format)
			continue;
		if (num_filters(i) < filternum)
			continue;
		/* success */
		j = 1;
		list_for_each_entry(fn, &s->fci->filters, node)
			if (filternum <= 0 || j++ == filternum)
				break;
		return fn;
	}
	return NULL;
}

static void start_stream_writer(int slot_num)
{
	int ret, fds[3] = {1, -1, -1};
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];
	char *glob = glob_cmd(a->write_cmd);

	PARA_INFO_LOG("starting stream writer: %s\n", glob? glob : a->write_cmd);
	open_filters(slot_num);

	ret = para_exec_cmdline_pid(&s->wpid, glob? glob : a->write_cmd, fds);
	free(glob);
	if (ret < 0) {
		PARA_ERROR_LOG("exec failed (%d)\n", ret);
		return;
	}
	s->write_fd = fds[0];
	add_close_on_fork_list(s->write_fd);
	/* we write to this fd in do_select, so we need non-blocking */
	fcntl(s->write_fd, F_SETFL, O_NONBLOCK);
	gettimeofday(&s->wstime, NULL);
	current_decoder = slot_num;
	activate_inactive_grab_clients(slot_num, s->format, &s->fci->filters);
}

static void open_receiver(int format)
{
	struct audio_format_info *a = &afi[format];
	struct slot_info *s;
	int ret, slot_num;

	slot_num = get_empty_slot();
	if (slot_num < 0)
		clean_exit(EXIT_FAILURE, PARA_STRERROR(-slot_num));
	s = &slot[slot_num];
	s->format = format;
	gettimeofday(&s->rtime, NULL);
	s->wtime = s->rtime;
	s->receiver_node = para_calloc(sizeof(struct receiver_node));
	s->receiver_node->conf = a->receiver_conf;
	ret = a->receiver->open(s->receiver_node);
	if (ret < 0) {
		PARA_ERROR_LOG("failed to open receiver (%s)\n",
			PARA_STRERROR(-ret));
		free(s->receiver_node);
		s->receiver_node = NULL;
		return;
	}
	PARA_NOTICE_LOG("started %s: %s receiver in slot %d\n",
		audio_formats[s->format], a->receiver->name, slot_num);
}

static int is_frozen(int format)
{
	struct timeval now;
	struct audio_format_info *a = &afi[format];

	gettimeofday(&now, NULL);
	return (tv_diff(&now, &a->restart_barrier, NULL) > 0)? 0 : 1;
}

static void start_current_receiver(void)
{
	int i;

	if (!af_status)
		return;
	i = get_audio_format_num(af_status);
	if (i < 0)
		return;
	if ((decoder_running(i) & 1) || is_frozen(i))
		return;
	open_receiver(i);
}

static void compute_time_diff(const struct timeval *status_time)
{
	struct timeval now, tmp, diff;
	static int count;
	int sign;
	const struct timeval max_deviation = {0, 500 * 1000};
	const int time_smooth = 5;

	gettimeofday(&now, NULL);
	sign = tv_diff(status_time, &now, &diff);
//		PARA_NOTICE_LOG("%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
	if (!count) {
		sa_time_diff_sign = sign;
		sa_time_diff = diff;
		count++;
		return;
	}
	if (count > 5) {
		int s = tv_diff(&diff, &sa_time_diff, &tmp);
		if (tv_diff(&max_deviation, &tmp, NULL) < 0)
			PARA_WARNING_LOG("time diff jump: %lims\n",
				s * tv2ms(&tmp));
	}
	count++;
	sa_time_diff_sign = tv_convex_combination(
		sa_time_diff_sign * time_smooth, &sa_time_diff,
		count > 10? sign : sign * time_smooth, &diff,
		&tmp);
	sa_time_diff = tmp;
	PARA_INFO_LOG("time diff (cur/avg): "
		"%li:%lu/%li:%lu\n",
		sign * diff.tv_sec, (diff.tv_usec + 500) / 1000,
		sa_time_diff_sign * sa_time_diff.tv_sec,
		(sa_time_diff.tv_usec + 500)/ 1000);
}

static void check_stat_line(char *line)
{
	int itemnum;
	size_t ilen = 0;
	struct timeval tv;

	if (!line)
		return;
	ringbuffer_add(stat_item_ringbuf, line);
	stat_client_write(line);
	itemnum = stat_line_valid(line);
	if (itemnum < 0)
		return;
	ilen = strlen(status_item_list[itemnum]);
	switch (itemnum) {
	case SI_STATUS:
		playing = strstr(line, "playing")? 1 : 0;
		break;
	case SI_FORMAT:
		free(af_status);
		af_status = para_strdup(line + ilen + 1);
		break;
	case SI_OFFSET:
		offset_seconds = atoi(line + ilen + 1);
		break;
	case SI_LENGTH:
		length_seconds = atoi(line + ilen + 1);
		break;
	case SI_STREAM_START:
		if (sscanf(line + ilen + 1, "%lu.%lu",
				&tv.tv_sec, &tv.tv_usec) == 2)
			server_stream_start = tv;
		break;
	case SI_CURRENT_TIME:
		if (sscanf(line + ilen + 1, "%lu.%lu", &tv.tv_sec,
				&tv.tv_usec) == 2)
			compute_time_diff(&tv);
		break;
	}
}

static void handle_signal(int sig)
{
	switch (sig) {
	case SIGCHLD:
		return check_sigchld();
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		PARA_EMERG_LOG("terminating on signal %d\n", sig);
		clean_exit(EXIT_FAILURE, "caught deadly signal");
		return;
	}
}

static void check_timeouts(void)
{
	struct timeval now;
	int slot_num, timeout = conf.stream_timeout_arg;

	gettimeofday(&now, NULL);
	FOR_EACH_SLOT(slot_num) {
		struct slot_info *s = &slot[slot_num];
		if (s->format < 0)
			continue;
		/* check read time */
		if (s->receiver_node &&
			now.tv_sec > s->rtime.tv_sec + timeout) {
			PARA_INFO_LOG("%s input buffer (slot %d) not ready\n",
				audio_formats[s->format], slot_num);
			if (s->fci)
				s->fci->error = 42;
			else
				close_receiver(slot_num);
		}
		/* check write time */
		if (s->wpid > 0 && !s->wkilled &&
			now.tv_sec > s->wtime.tv_sec + timeout) {
			PARA_INFO_LOG("%s output buffer (slot %d) not ready\n",
				audio_formats[s->format], slot_num);
			if (s->fci)
				s->fci->error = 42;
		}
	}
}

static size_t get_loaded_bytes(int slot_num)
{
	size_t loaded = 0;
	struct slot_info *s = &slot[slot_num];
	struct receiver_node *rn = s->receiver_node;

	if (s->format < 0)
		goto out;

	if (afi[s->format].num_filters) {
		if (s->fci)
			loaded = *s->fci->out_loaded;
	} else {
		if (rn)
			loaded = rn->loaded;
	}
out:
	return loaded;
}


static void close_decoder_if_idle(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct receiver_node *rn = s->receiver_node;

	if (s->format < 0)
		return;
	if (!s->fci)
		return;
	if (!rn->eof && !s->fci->error && s->wpid > 0)
		return;
	if (!s->fci->error && s->wpid > 0) { /* eof */
		if (filter_io(s->fci) > 0)
			return;
		if (get_loaded_bytes(slot_num))
			return;
	}
	if (s->write_fd > 0) {
		PARA_INFO_LOG("slot %d: closing write fd %d\n", slot_num,
			s->write_fd);
		close(s->write_fd);
		del_close_on_fork_list(s->write_fd);
		s->write_fd = -1;
	}
	if (s->wpid > 0)
		return; /* wait until writer dies before closing filters */
	PARA_INFO_LOG("closing all filters in slot %d (filter_chain %p)\n",
		slot_num, s->fci);
	close_filters(s->fci);
	free(s->fci);
	close_receiver(slot_num);
	clear_slot(slot_num);
}

static int set_stream_fds(fd_set *wfds)
{
	int i, max_fileno = -1;

	check_timeouts();
	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct audio_format_info *a;
		struct receiver_node *rn;

		close_decoder_if_idle(i);
		s->wcheck = 0;
		if (s->format < 0)
			continue;
		a = &afi[s->format];
		rn = s->receiver_node;
		if (rn && rn->loaded && !s->wpid) {
			PARA_INFO_LOG("no writer in slot %d\n", i);
			start_stream_writer(i);
		}
		if (s->write_fd <= 0)
			continue;
		if (!get_loaded_bytes(i))
			continue;
		FD_SET(s->write_fd, wfds);
		s->wcheck = 1;
		max_fileno = MAX(s->write_fd, max_fileno);
	}
//	PARA_INFO_LOG("return %d\n", max_fileno);
	return max_fileno;
}

static int write_audio_data(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = &afi[s->format];
	struct receiver_node *rn = s->receiver_node;
	int rv;
	char **buf;
	size_t *len;

	if (a->num_filters) {
		buf = &s->fci->outbuf;
		len = s->fci->out_loaded;
	} else {
		buf = &rn->buf;
		len = &rn->loaded;
	}
	PARA_DEBUG_LOG("writing %p (%d bytes)\n", *buf, *len);
	rv = write(s->write_fd, *buf, *len);
	PARA_DEBUG_LOG("wrote %d/%d\n", rv, *len);
	if (rv < 0) {
		PARA_WARNING_LOG("write error in slot %d (fd %d): %s\n",
			slot_num, s->write_fd, strerror(errno));
		*len = 0;
		s->fci->error = E_WRITE_AUDIO_DATA;
	} else if (rv != *len) {
		PARA_DEBUG_LOG("partial %s write (%i/%i) for slot %d\n",
			audio_formats[s->format], rv, *len, slot_num);
		*len -= rv;
		memmove(*buf, *buf + rv, *len);
	} else
		*len = 0;
	if (rv > 0)
		gettimeofday(&s->wtime, NULL);
	return rv;
}

static void slot_io(fd_set *wfds)
{
	int ret, i;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct receiver_node *rn = s->receiver_node;

		if (rn && rn->loaded)
			gettimeofday(&s->rtime, NULL);
		if (s->format >= 0 && s->write_fd > 0 && s->fci) {
			ret = filter_io(s->fci);
			if (ret < 0)
				s->fci->error = -ret;
//			PARA_DEBUG_LOG("slot %d, filter io %d bytes, check write: %d, loaded: %d/%d, eof: %d\n",
//				 i, ret, s->wcheck, rn->loaded, *s->fci->out_loaded, rn->eof);
		}
		if (s->write_fd <= 0 || !s->wcheck || !FD_ISSET(s->write_fd, wfds))
			continue;
		write_audio_data(i);
	}
}

static int parse_stream_command(const char *txt, char **cmd)
{
	char *p = strchr(txt, ':');
	int i;

	if (!p)
		return -E_MISSING_COLON;
	p++;
	FOR_EACH_AUDIO_FORMAT(i) {
		if (strncmp(txt, audio_formats[i], strlen(audio_formats[i])))
			continue;
		*cmd = p;
		return i;
	}
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

static int add_filter(int format, char *cmdline)
{
	struct audio_format_info *a = &afi[format];
	int filter_num, nf = a->num_filters;

	filter_num = check_filter_arg(cmdline, &a->filter_conf[nf]);
	if (filter_num < 0)
		return filter_num;
	a->filters[nf] = &filters[filter_num];
	a->num_filters++;
	PARA_INFO_LOG("%s filter %d: %s\n", audio_formats[format], nf + 1,
		a->filters[nf]->name);
	return filter_num;
}

static int setup_default_filters(void)
{
	int i, ret = 1;

	FOR_EACH_AUDIO_FORMAT(i) {
		struct audio_format_info *a = &afi[i];
		char *tmp;
		int j;
		if (a->num_filters)
			continue;
		/* add "dec" to audio format name */
		tmp = make_message("%sdec", audio_formats[i]);
		for (j = 0; filters[j].name; j++)
			if (!strcmp(tmp, filters[j].name))
				break;
		free(tmp);
		ret = -E_UNSUPPORTED_FILTER;
		if (!filters[j].name)
			goto out;
		tmp = para_strdup(filters[j].name);
		ret = add_filter(i, tmp);
		free(tmp);
		if (ret < 0)
			goto out;
		PARA_INFO_LOG("%s -> default filter: %s\n", audio_formats[i], filters[j].name);
		ret = add_filter(i, "wav");
		if (ret < 0)
			goto out;
		PARA_INFO_LOG("%s -> default filter: wav\n", audio_formats[i]);
	}
out:
	return ret;
}

static int init_stream_io(void)
{
	int i, ret, receiver_num;
	char *cmd;

	for (i = 0; i < conf.stream_write_cmd_given; i++) {
		ret = parse_stream_command(conf.stream_write_cmd_arg[i], &cmd);
		if (ret < 0)
			goto out;
		afi[ret].write_cmd = para_strdup(cmd);
		PARA_INFO_LOG("%s write command: %s\n", audio_formats[ret], afi[ret].write_cmd);
	}
	for (i = 0; receivers[i].name; i++) {
		PARA_INFO_LOG("initializing %s receiver\n", receivers[i].name);
		receivers[i].init(&receivers[i]);
	}
	for (i = 0; i < conf.receiver_given; i++) {
		char *arg = conf.receiver_arg[i];
		char *recv = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!recv)
			goto out;
		*recv = '\0';
		recv++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		afi[ret].receiver_conf = check_receiver_arg(recv, &receiver_num);
		if (!afi[ret].receiver_conf) {
			ret = -E_RECV_SYNTAX;
			goto out;
		}
		afi[ret].receiver = &receivers[receiver_num];
	}
	/* use the first available receiver with no arguments
	 * for those audio formats for which no receiver
	 * was specified
	 */
	cmd = para_strdup(receivers[0].name);
	FOR_EACH_AUDIO_FORMAT(i) {
		struct audio_format_info *a = &afi[i];
		if (a->receiver_conf)
			continue;
		a->receiver_conf = check_receiver_arg(cmd, &receiver_num);
		if (!a->receiver_conf)
			return -E_RECV_SYNTAX;
		a->receiver = &receivers[receiver_num];
	}
	free(cmd);
	/* filters */
	filter_init(filters);
	FOR_EACH_AUDIO_FORMAT(i) {
		afi[i].filter_conf = para_malloc((conf.filter_given + 1) * sizeof(char *));
		afi[i].filters = para_malloc((conf.filter_given + 1) * sizeof(struct filter *));
	}
	if (!conf.no_default_filters_given)
		return setup_default_filters();
	for (i = 0; i < conf.filter_given; i++) {
		char *arg = conf.filter_arg[i];
		char *filter_name = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!filter_name)
			goto out;
		*filter_name = '\0';
		filter_name++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		ret = add_filter(ret, filter_name);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	return ret;
}

static int dump_commands(int fd)
{
	char *buf = para_strdup(""), *tmp = NULL;
	int i;
	ssize_t ret;

	FOR_EACH_COMMAND(i) {
		tmp = make_message("%s%s\t%s\n", buf, cmds[i].name,
			cmds[i].description);
		free(buf);
		buf = tmp;
	}
	ret = client_write(fd, buf);
	free(buf);
	return ret;
}

/*
 * command handlers don't close their fd on errors (ret < 0) so that
 * its caller can send an error message. Otherwise (ret >= 0) it's up
 * to each individual command to close the fd if necessary.  
 */

static int com_help(int fd, int argc, char **argv)
{
	int i, ret;
	char *buf;
	const char *dflt = "No such command. Available commands:\n";

	if (argc < 2) {
		ret = dump_commands(fd);
		goto out;
	}
	FOR_EACH_COMMAND(i) {
		if (strcmp(cmds[i].name, argv[1]))
			continue;
		buf = make_message(
			"NAME\n\t%s -- %s\n"
			"SYNOPSIS\n\tpara_audioc %s\n"
			"DESCRIPTION\n%s\n",
			argv[1],
			cmds[i].description,
			cmds[i].synopsis,
			cmds[i].help
		);
		ret = client_write(fd, buf);
		free(buf);
		goto out;
	}
	ret = client_write(fd, dflt);
	if (ret > 0)
		ret = dump_commands(fd);
out:
	if (ret >= 0)
		close(fd);
	return ret;
}

static int com_stat(int fd, __unused int argc, __unused char **argv)
{
	int i, ret;
	char *buf = audiod_status_string();

	buf = para_strcat(buf, "\n");
	for (i = RINGBUFFER_SIZE - 1; i >= 0; i--) {
		char *tmp, *line = ringbuffer_get(stat_item_ringbuf, i);
		if (!line)
			continue;
		tmp = make_message("%s\n", line);
		buf = para_strcat(buf, tmp);
		free(tmp);
	}
	ret = client_write(fd, buf);
	if (ret > 0)
		ret = stat_client_add(fd);
	free(buf);
	return ret;
}

#if 0
static char *list_filters(void)
{
	int i, j;
	char *tmp, *msg = make_message("format\tnum\tcmd\n");

	FOR_EACH_AUDIO_FORMAT(i) {
		for (j = 0; j < afi[i].num_filters; j++) {
			tmp = make_message("%s\t%i\t%s\n",
				afi[i].name, j, afi[i].filter_cmds[j]);
			msg = para_strcat(msg, tmp);
			free(tmp);
		}
		tmp = make_message("%s\t%i\t%s\n", afi[i].name,
			j, afi[i].write_cmd);
		msg = para_strcat(msg, tmp);
		free(tmp);
	}
	return msg;
}
#endif

static int com_grab(int fd, int argc, char **argv)
{
	struct grab_client *gc;
	struct filter_node *fn;
	int err;

	PARA_INFO_LOG("argc: %d, argv[0]: %s, optind: %d\n", argc, argv[0], optind);
	gc = grab_client_new(fd, argc, argv, &err);
	PARA_INFO_LOG("argc: %d, argv[0]: %s, optind: %d\n", argc, argv[0], optind);
	if (!gc)
		goto err_out;
	fn = find_filter_node(gc->conf->slot_arg, gc->audio_format_num, gc->conf->filter_num_arg);
	if (fn)
		activate_grab_client(gc, fn);
	return 1;
err_out:
	if (err != -E_GC_HELP_GIVEN)
		return err;
	err = client_write(fd, "Usage: para_audioc [audioc_options] -- "
		"grab [grab_options]\nAvailable options:\n");
	if (err < 0)
		return err;
	err = client_write(fd, GRAB_HELP_TXT);
	if (err < 0)
		return err;
	close(fd);
	return 1;
}

static int __noreturn com_term(int fd, __unused int argc, __unused char **argv)
{
	close(fd);
	clean_exit(EXIT_SUCCESS, "terminating on user request");
}

static int com_on(int fd, __unused int argc, __unused char **argv)
{
	audiod_status = AUDIOD_ON;
	close(fd);
	return 1;
}

static int com_off(int fd, __unused int argc, __unused char **argv)
{
	audiod_status = AUDIOD_OFF;
	close(fd);
	return 1;
}

static int com_sb(int fd, __unused int argc, __unused char **argv)
{
	audiod_status = AUDIOD_STANDBY;
	close(fd);
	return 1;
}

static int com_cycle(int fd, int argc, char **argv)
{
	switch (audiod_status) {
		case  AUDIOD_ON:
			return com_sb(fd, argc, argv);
			break;
		case  AUDIOD_OFF:
			return com_on(fd, argc, argv);
			break;
		case  AUDIOD_STANDBY:
			return com_off(fd, argc, argv);
			break;
	}
	close(fd);
	return 1;
}

static int check_perms(struct ucred *c)
{
	int i;

	if (!conf.user_allow_given)
		return 1;
	for (i = 0; i < conf.user_allow_given; i++)
		if (c->uid == conf.user_allow_arg[i])
			return 1;
	return -E_UCRED_PERM;
}

static int handle_connect(void)
{
	int i, argc, ret, clifd = -1;
	struct ucred c;
	char *buf = para_malloc(MAXLINE), **argv = NULL;
	struct sockaddr_un unix_addr;

	ret = para_accept(audiod_socket, &unix_addr, sizeof(struct sockaddr_un));
	if (ret < 0)
		goto out;
	clifd = ret;
	ret = recv_cred_buffer(clifd, buf, MAXLINE - 1, &c);
	if (ret < 0)
		goto out;
	PARA_INFO_LOG("pid: %i, uid: %i, gid: %i, ret: %i, buf: %s\n", c.pid, c.uid, c.gid, ret, buf);
	buf[ret] = '\0';
	ret = check_perms(&c);
	if (ret < 0)
		goto out;
	argc = split_args(buf, &argv, '\n');
	PARA_INFO_LOG("argv[0]: %s\n", argv[0]);
	for (i = 0; cmds[i].name; i++) {
		if (strcmp(cmds[i].name, argv[0]))
			continue;
		ret = cmds[i].handler(clifd, argc + 1, argv);
		goto out;
	}
	ret = -E_INVALID_AUDIOD_CMD; /* cmd not found */
out:
	free(buf);
	free(argv);
	if (clifd > 0 && ret < 0 && ret != -E_CLIENT_WRITE) {
		char *tmp = make_message("%s\n", PARA_STRERROR(-ret));
		client_write(clifd, tmp);
		free(tmp);
		close(clifd);
	}
	return ret;
}

static void audiod_get_socket(void)
{
	struct sockaddr_un unix_addr;

	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	PARA_NOTICE_LOG("connecting to local socket %s\n", socket_name);
	if (conf.force_given)
		unlink(socket_name);
	audiod_socket = create_pf_socket(socket_name, &unix_addr,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (audiod_socket < 0) {
		PARA_EMERG_LOG("%s", "can not connect to socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	if (listen(audiod_socket, 5) < 0) {
		PARA_EMERG_LOG("%s", "can not listen on socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	add_close_on_fork_list(audiod_socket);
}

static int open_stat_pipe(void)
{
	int ret, fd[3] = {-1, 1, 0};
	char *argv[] = {BINDIR "/para_client", "stat", NULL};
	pid_t pid;
	ret = para_exec(&pid, BINDIR "/para_client",  argv, fd);
	if (ret >= 0) {
		ret = fd[1];
		PARA_NOTICE_LOG("stat pipe opened, fd %d\n", ret);
		add_close_on_fork_list(ret);
	} else
		clean_exit(EXIT_FAILURE, "failed to open status pipe");
	return ret;
}

static int pre_select(fd_set *rfds, fd_set *wfds, struct timeval *tv)
{
	int i, ret, max = -1;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct audio_format_info *a;
		struct receiver_node *rn = s->receiver_node;
		if (s->format < 0 || !rn)
			continue;
		a = &afi[s->format];
		ret = a->receiver->pre_select(rn, rfds, wfds, tv);
//		PARA_NOTICE_LOG("%s preselect: %d\n", a->receiver->name, ret);
		max = MAX(max, ret);
	}
	return max;

}
static void audiod_post_select(int select_ret, fd_set *rfds, fd_set *wfds)
{
	int i, ret;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct audio_format_info *a;
		struct receiver_node *rn = s->receiver_node;
		if (s->format < 0 || !rn || rn->eof)
			continue;
		a = &afi[s->format];
		ret = a->receiver->post_select(rn, select_ret, rfds, wfds);
		if (ret <= 0) {
			if (ret)
				PARA_ERROR_LOG("%s post select failed: %s (slot %d)\n",
				a->receiver->name, PARA_STRERROR(-ret), i);
			else
				PARA_INFO_LOG("eof in slot %d\n", i);
			rn->eof = 1;
		}
		if (ret < 0 && s->fci)
			s->fci->error = ret;
	}
}

/* TODO: move everything before the select call to pre_select() */
static void __noreturn audiod_mainloop(void)
{
	fd_set rfds, wfds;
	int ret, max_fileno, sbo = 0;
	char status_buf[STRINGSIZE] = "";
	struct timeval tv;
repeat:
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = 0;
	if (audiod_status != AUDIOD_ON)
		kill_all_decoders();
	else if (playing)
		start_current_receiver();
	max_fileno = set_stream_fds(&wfds);
	/* stat pipe (read) */
	if (stat_pipe >= 0 && audiod_status == AUDIOD_OFF)
		close_stat_pipe();
	if (stat_pipe < 0 && audiod_status != AUDIOD_OFF) {
		stat_pipe = open_stat_pipe();
		sbo = 0;
		status_buf[0] = '\0';
	}
	if (stat_pipe >= 0 && audiod_status != AUDIOD_OFF) {
		FD_SET(stat_pipe, &rfds);
		max_fileno = MAX(max_fileno, stat_pipe);
	}
	/* always check signal pipe */
	FD_SET(signal_pipe, &rfds);
	max_fileno = MAX(max_fileno, signal_pipe);
	/* local socket */
	if (audiod_socket < 0)
		audiod_get_socket(); /* doesn't return on errors */
	FD_SET(audiod_socket, &rfds);
	max_fileno = MAX(max_fileno, audiod_socket);
	tv.tv_sec = 0;
	tv.tv_usec = 200 * 1000;
	ret = pre_select(&rfds, &wfds, &tv);
	max_fileno = MAX(max_fileno, ret);
	ret = select(max_fileno + 1, &rfds, &wfds, NULL, &tv);
	if (ret < 0 && errno != EINTR)
		PARA_ERROR_LOG("select returned %d (%s)\n", ret,
			strerror(errno));
	if (audiod_status != AUDIOD_OFF)
		audiod_status_dump();
	if (ret < 0)
		goto repeat;
	audiod_post_select(ret, &rfds, &wfds);
	/* read status pipe */
	if (stat_pipe >=0 && FD_ISSET(stat_pipe, &rfds)) {
		ret = read(stat_pipe, status_buf + sbo, STRINGSIZE - 1 - sbo);
		if (ret <= 0) {
			close_stat_pipe();
			/* avoid busy loop if server is down */
			while (sleep(1) > 0)
				; /* try again*/
		} else {
			status_buf[ret + sbo] = '\0';
			sbo = for_each_line(status_buf, ret + sbo,
				&check_stat_line, 0);
		}
	}
	slot_io(&wfds);
	if (FD_ISSET(audiod_socket, &rfds)) {
		ret = handle_connect();
		if (ret < 0) {
			PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		}
	}
	/* signals */
	if (FD_ISSET(signal_pipe, &rfds)) {
		int sig_nr = para_next_signal();
		if (sig_nr > 0)
			handle_signal(sig_nr);
	}
	goto repeat;
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		PARA_WARNING_LOG("%s", "invalid mode\n");
}

int __noreturn main(int argc, char *argv[])
{
	char *cf;
	int i;

	valid_fd_012();
	hostname = para_hostname();
	cmdline_parser(argc, argv, &conf);
	para_drop_privileges(conf.user_arg);
	cf = configfile_exists();
	if (cf) {
		if (cmdline_parser_configfile(cf, &conf, 0, 0, 0)) {
			fprintf(stderr, "parse error in config file\n");
			exit(EXIT_FAILURE);
		}
	}
	log_welcome("para_audiod", conf.loglevel_arg);
	i = init_stream_io();
	if (i < 0) {
		fprintf(stderr, "init stream io error: %s\n",
			PARA_STRERROR(-i));
		exit(EXIT_FAILURE);
	}
	server_uptime(UPTIME_SET);
	set_initial_status();
	FOR_EACH_SLOT(i)
		clear_slot(i);
	stat_item_ringbuf = ringbuffer_new(RINGBUFFER_SIZE);
	init_grabbing();
	setup_signal_handling();
	if (conf.daemon_given)
		daemon_init();
	audiod_mainloop();
}
